@@ -74,7 +74,7 @@ function getChildData(child) {
  *                                    - childrenTransitions
  *                                    - nodesShouldEnter
  */
-function getInitialTransitionState(oldChildren, nextChildren) {
+export function getInitialTransitionState(oldChildren, nextChildren) {
   let nodesWillExit = false;
   let nodesWillEnter = false;
 
@@ -82,7 +82,6 @@ function getInitialTransitionState(oldChildren, nextChildren) {
     if (!newChild || oldChild.type !== newChild.type) {
       return {};
     }
-
     const { entering, exiting } =
       getNodeTransitions(getChildData(oldChild), getChildData(newChild)) || {};
 
@@ -257,7 +256,7 @@ function getChildPropsOnEnter(animate, data, enteringNodes, cb) {
  *
  * @return {Function}              Child-prop transformation function.
  */
-function getTransitionPropsFactory(props, state, setState) {
+export function getTransitionPropsFactory(props, state, setState) {
   const nodesWillExit = state && state.nodesWillExit;
   const nodesWillEnter = state && state.nodesWillEnter;
   const nodesShouldEnter = state && state.nodesShouldEnter;
@@ -402,8 +401,3 @@ function getTransitionPropsFactory(props, state, setState) {
     return { animate, data };
   };
 }
-
-export default {
-  getInitialTransitionState,
-  getTransitionPropsFactory
-};
