@@ -1,7 +1,7 @@
 import { expectAssignable } from 'tsd'
-import Undici, { Pool, Client, Errors } from '../..'
+import Undici, { Pool, Client, errors } from '../..'
 
 expectAssignable<Pool>(Undici('', {}))
 expectAssignable<Pool>(new Undici.Pool('', {}))
 expectAssignable<Client>(new Undici.Client('', {}))
-expectAssignable<typeof Errors>(Undici.errors)
+expectAssignable<typeof errors>(Undici.errors)
